/* Copyright (c) 2022, Arm Limited
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stddef.h>

#include "hal/dma_api.h"

#include "dma350_lib.h"
#include "mps3_dma.h"
#include "SSE310MPS3.h"

#ifndef container_of
// Returns a pointer to the `type` containing the `member` pointed by `ptr`;
#define container_of(ptr, type, member)                    \
    ({                                                     \
        const typeof(((type *)0)->member) *__mptr = (ptr); \
        (type *)((char *)__mptr - offsetof(type, member)); \
    })
#endif

typedef struct priv_mps3_dma_s {
    mps3_dma_t mps3_dma;
    struct dma350_ch_dev_t *device;

    mdh_dma_callback_f on_event;
    void *context;
} priv_mps3_dma_t;

static mdh_dma_error_t
dma_memmove(const mdh_dma_t *self, void *src, void *des, uint32_t size, mdh_dma_callback_f cbk, void *ctx);

static mdh_dma_state_t get_state(const mdh_dma_t *self);

static priv_mps3_dma_t *get_priv_mps3_dma_instance(const struct dma350_ch_dev_t *dma350_instance);

static bool does_memory_overlap(void *src, void *des, uint32_t size);

static const mdh_dma_vtable_t gsc_vtable = {
    .dma_memmove = dma_memmove,
    .get_state = get_state,
};

#ifdef DMA350_CH0_S
static priv_mps3_dma_t gs_priv_dma_0 = {.mps3_dma.dma.vfptr = &gsc_vtable, .device = &DMA350_DMA0_CH0_DEV_S};

void DMA_Channel_0_Handler(void)
{
    dma350_clear_done_irq(gs_priv_dma_0.device);
    if (gs_priv_dma_0.on_event != NULL) {
        gs_priv_dma_0.on_event(gs_priv_dma_0.context);
    }
}
#endif

#ifdef DMA350_CH1_S
static priv_mps3_dma_t gs_priv_dma_1 = {.mps3_dma.dma.vfptr = &gsc_vtable, .device = &DMA350_DMA0_CH1_DEV_S};

void DMA_Channel_1_Handler(void)
{
    dma350_clear_done_irq(gs_priv_dma_1.device);
    if (gs_priv_dma_1.on_event != NULL) {
        gs_priv_dma_1.on_event(gs_priv_dma_1.context);
    }
}
#endif

static mdh_dma_error_t
dma_memmove(const mdh_dma_t *self, void *src, void *des, uint32_t size, mdh_dma_callback_f cbk, void *ctx)
{
    priv_mps3_dma_t *that = container_of(self, priv_mps3_dma_t, mps3_dma.dma);
    enum dma350_lib_error_t ret = DMA350_LIB_ERR_NONE;

    if (get_state(self) != MDH_DMA_STATE_READY) {
        return MDH_DMA_ERROR_NOT_READY;
    }

    if (cbk != NULL) {
        /* Enable NVIC for the channel */
        NVIC_EnableIRQ(DMA_CHANNEL_0_IRQn + that->device->cfg.channel);

        that->on_event = cbk;
        that->context = ctx;
    } else {
        /* Disable NVIC for the channel */
        NVIC_DisableIRQ(DMA_CHANNEL_0_IRQn + that->device->cfg.channel);

        that->on_event = NULL;
        that->context = NULL;
    }

    if (does_memory_overlap(src, des, size)) {
        ret = dma350_memmove(that->device, src, des, size, DMA350_LIB_EXEC_IRQ);
    } else {
        ret = dma350_memcpy(that->device, src, des, size, DMA350_LIB_EXEC_IRQ);
    }

    if (DMA350_LIB_ERR_NONE != ret) {
        return MDH_DMA_ERROR_INTERNAL_ERROR;
    }

    return MDH_DMA_ERROR_NO_ERROR;
}

static mdh_dma_state_t get_state(const mdh_dma_t *self)
{
    priv_mps3_dma_t *that = container_of(self, priv_mps3_dma_t, mps3_dma.dma);
    union dma350_ch_status_t status;

    dma350_get_status(that->device, &status);

    if (1 == status.b.STAT_ERR) {
        return MDH_DMA_STATE_ERROR;
    } else if (dma350_ch_is_busy(that->device)) {
        return MDH_DMA_STATE_BUSY;
    } else {
        return MDH_DMA_STATE_READY;
    }
}

static priv_mps3_dma_t *get_priv_mps3_dma_instance(const struct dma350_ch_dev_t *dma350_instance)
{
#ifdef DMA350_CH0_S
    if (dma350_instance == &DMA350_DMA0_CH0_DEV_S) {
        return &gs_priv_dma_0;
    }
#endif // DMA350_CH0_NS

#ifdef DMA350_CH1_S
    if (dma350_instance == &DMA350_DMA0_CH1_DEV_S) {
        return &gs_priv_dma_1;
    }
#endif // DMA350_CH1_NS

    return NULL;
}

bool mps3_dma_init(mps3_dma_t **dma, struct dma350_ch_dev_t *dma350_instance)
{
    priv_mps3_dma_t *priv_dma = get_priv_mps3_dma_instance(dma350_instance);
    if (NULL == priv_dma) {
        return false;
    }

    *dma = &(priv_dma->mps3_dma);

    return true;
}

static bool does_memory_overlap(void *src, void *des, uint32_t size)
{
    if (((src < des) && (((uint8_t *)src + size) < (uint8_t *)des))
        || ((des < src) && (((uint8_t *)des + size) < (uint8_t *)src))) {
        return false;
    } else {
        return true;
    }
}
